import logging
from typing import Iterable, Optional, Tuple, Union

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, FactMetadata, FactStatus, FactType

ConceptValueType = str  # TODO: describe concept type

logger = logging.getLogger(__name__)


class ConceptFact(AbstractFact[ConceptValueType]):
    def __init__(self, id_: Optional[str], status: FactStatus, type_id: str,
                 value: Union[ConceptValueType, Iterable[ConceptValueType]] = tuple(),
                 mention: Optional[Iterable[AbstractTalismanSpan]] = None,
                 metadata: Optional[FactMetadata] = None):
        if value is None:
            value = tuple()
            logger.warning("value must not be None, value=tuple()")
        elif isinstance(value, ConceptValueType):
            pass  # do nothing
        elif isinstance(value, Iterable):
            value = tuple(value)
        else:
            raise ValueError(f"Bad concept fact value {value}")
        super().__init__(id_, FactType.CONCEPT, status, type_id, value, mention, metadata)

    def with_changes(self: 'ConceptFact', *, status: FactStatus = None, type_id: str = None,
                     value: Union[ConceptValueType, Tuple[ConceptValueType, ...]] = None,
                     mention: Tuple[AbstractTalismanSpan, ...] = None,
                     metadata: FactMetadata = None) -> 'ConceptFact':
        return ConceptFact(
            self._id,
            status if status is not None else self._status,
            type_id if type_id is not None else self._type_id,
            value if value is not None else self._value,
            mention if mention is not None else self._mention,
            metadata if metadata is not None else self._metadata
        )



from typing import Any, Callable, Dict, Tuple

from typing_extensions import Literal

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, FactMetadata, FactStatus, FactType
from tdm.datamodel.fact import ConceptFact
from tdm.json_schema.fact.common import AbstractFactModel


class ConceptFactModel(AbstractFactModel):
    fact_type: Literal[FactType.CONCEPT] = FactType.CONCEPT

    def to_value(self, mapping: Dict[str, AbstractFact]) -> Any:
        return self.value

    @property
    def fact_factory(self) -> Callable[[str, FactStatus, str, Any, Tuple[AbstractTalismanSpan, ...], FactMetadata],
                                       ConceptFact]:
        return ConceptFact

    @classmethod
    def build_value(cls, value: Any) -> Any:
        return value



from typing import Optional, Tuple

from tdm.abstract.json_schema import AbstractContentModel
from tdm.datamodel import TreeDocumentContent
from tdm.json_schema.content.markup import NodeMarkup
from tdm.json_schema.content.metadata import NodeMetadata


class TreeDocumentContentModel(AbstractContentModel):
    id: str
    metadata: NodeMetadata
    text: str
    nodes: Optional[Tuple['TreeDocumentContentModel', ...]]
    markup: NodeMarkup = NodeMarkup()

    def to_content(self) -> TreeDocumentContent:
        return self.markup.to_doc(
            node_id=self.id,
            text=self.text,
            nodes=(node.to_content() for node in self.nodes) if self.nodes is not None else None,
            **self.metadata.parse()
        )

    @classmethod
    def build(cls, doc: TreeDocumentContent) -> 'TreeDocumentContentModel':
        metadata_type = cls.__fields__['metadata'].type_
        markup_type = cls.__fields__['markup'].type_

        return TreeDocumentContentModel(
            id=doc.id,
            metadata=metadata_type.build(doc),
            text=doc.node_text,
            nodes=tuple(cls.build(node) for node in doc.nodes) if doc.nodes is not None else None,
            markup=markup_type.build(doc)
        )


TreeDocumentContentModel.update_forward_refs()



from typing import TypeVar

from tdm.abstract.datamodel.directive import AbstractDirective, DirectiveType
from tdm.json_schema.directive import CreateConceptDirectiveModel
from tdm.json_schema.directive.abstract import AbstractDirectiveModel


_AbstractDirective = TypeVar('_AbstractDirective', bound=AbstractDirective)


_DIRECTIVE_MODELS = {
    DirectiveType.CREATE_CONCEPT: CreateConceptDirectiveModel
}


def build_directive_model(directive: _AbstractDirective) -> AbstractDirectiveModel:
    return _DIRECTIVE_MODELS[directive.directive_type].build(directive)



__all__ = [
    'TreeDocumentContent',
    'TalismanDocument',
    'CreateConceptDirective',
    'ConceptFact', 'PropertyFact', 'PropertyLinkValue', 'RelationFact', 'RelationLinkValue', 'ValueFact',
    'DefaultSpan', 'NullableSpan', 'TalismanSpan'
]

from .content import TreeDocumentContent
from .directive.create_concept import CreateConceptDirective
from .document import TalismanDocument
from .fact import ConceptFact, PropertyFact, PropertyLinkValue, RelationFact, RelationLinkValue, ValueFact
from .span import DefaultSpan, NullableSpan, TalismanSpan



from typing import Iterable, Optional

from pydantic import BaseModel

from tdm.datamodel import TreeDocumentContent


class NodeMarkup(BaseModel):
    class Config:
        extra = 'allow'  # any other extra fields will be kept

    def to_doc(self, node_id: str, text: str, nodes: Optional[Iterable[TreeDocumentContent]], **kwargs) -> TreeDocumentContent:
        return TreeDocumentContent(
            node_id=node_id,
            text=text,
            nodes=nodes,
            markup=self.dict(),
            **kwargs
        )

    @classmethod
    def build(cls, doc: TreeDocumentContent) -> 'NodeMarkup':
        return cls.construct(**doc.markup)



from typing import Any, Callable, Dict, Tuple

from typing_extensions import Literal

from tdm.abstract.datamodel import AbstractFact, AbstractTalismanSpan, FactMetadata, FactStatus, FactType
from tdm.datamodel.fact import RelationFact, RelationLinkValue
from tdm.json_schema.fact.common import AbstractFactModel, PropertyLinkValueModel


class RelationFactModel(AbstractFactModel):
    fact_type: Literal[FactType.RELATION] = FactType.RELATION
    value: PropertyLinkValueModel

    def to_value(self, mapping: Dict[str, AbstractFact]) -> RelationLinkValue:
        return RelationLinkValue(
            property_id=self.value.property_id,
            from_fact=mapping[self.value.from_fact],
            to_fact=mapping[self.value.to_fact]
        )

    @property
    def fact_factory(self) -> Callable[[str, FactStatus, str, Any, Tuple[AbstractTalismanSpan, ...], FactMetadata],
                                       AbstractFact]:
        return RelationFact

    @classmethod
    def build_value(cls, value: RelationLinkValue) -> PropertyLinkValueModel:
        return PropertyLinkValueModel.construct(property_id=value.property_id, from_fact=value.from_fact.id, to_fact=value.to_fact.id)



from abc import ABCMeta, abstractmethod
from enum import Enum
from operator import attrgetter
from typing import Any, Dict, Set, Tuple, TypeVar

from .span import AbstractSpan, AbstractTalismanSpan
from .tree import AbstractTree


class NodeType(Enum):
    HEADER = "header"
    TEXT = "text"
    LIST = "list"
    JSON = "json"
    KEY = "key"
    TABLE = "table"
    TABLE_ROW = "row"
    IMAGE = "image"


_AbstractDocumentContent = TypeVar('_AbstractDocumentContent', bound='AbstractDocumentContent')


class AbstractDocumentContent(metaclass=ABCMeta):
    __slots__ = ()

    @property
    @abstractmethod
    def type(self) -> NodeType:
        pass

    @property
    @abstractmethod
    def text(self) -> str:
        pass

    @abstractmethod
    def text_for(self, span: AbstractSpan) -> str:
        pass

    @property
    @abstractmethod
    def node_text(self) -> str:
        pass

    @abstractmethod
    def with_node_text(self: _AbstractDocumentContent, text: str) -> _AbstractDocumentContent:
        pass

    @abstractmethod
    def get_node_text_translation(self, lang: str) -> str:
        pass

    @property
    @abstractmethod
    def node_text_languages(self) -> Set[str]:
        pass

    @abstractmethod
    def with_node_translation(self: _AbstractDocumentContent, lang: str, translation: str) -> _AbstractDocumentContent:
        pass

    @abstractmethod
    def without_node_translation(self: _AbstractDocumentContent, lang: str) -> _AbstractDocumentContent:
        pass

    @property
    @abstractmethod
    def original_node_text(self) -> str:
        pass

    @property
    @abstractmethod
    def markup(self) -> Dict[str, Any]:
        pass

    @property
    @abstractmethod
    def is_hidden(self) -> bool:
        pass

    @abstractmethod
    def hide(self: _AbstractDocumentContent) -> _AbstractDocumentContent:
        pass

    @abstractmethod
    def show(self: _AbstractDocumentContent) -> _AbstractDocumentContent:
        pass


_AbstractTreeDocumentContent = TypeVar('_AbstractTreeDocumentContent', bound='AbstractTreeDocumentContent')


class AbstractTreeDocumentContent(AbstractTree, AbstractDocumentContent, metaclass=ABCMeta):

    @abstractmethod
    def talisman_spans(self, span: AbstractSpan) -> Tuple[AbstractTalismanSpan, ...]:
        pass

    def talisman_span(self, span: AbstractSpan) -> AbstractTalismanSpan:
        spans = self.talisman_spans(span)
        if len(spans) != 1:
            raise ValueError(f"{span} spreads over several nodes {tuple(map(attrgetter('node_id'), spans))}")
        return spans[0]

    @abstractmethod
    def span(self, talisman_span: AbstractTalismanSpan) -> AbstractSpan:
        pass



__all__ = [
    'ConceptFact',
    'PropertyFact', 'PropertyLinkValue',
    'RelationFact', 'RelationLinkValue',
    'ValueFact'
]

from .concept import ConceptFact
from .property import PropertyFact, PropertyLinkValue
from .relation import RelationFact, RelationLinkValue
from .value import ValueFact



from copy import deepcopy
from typing import Any, Callable, Dict, Iterable, Optional, Set, Tuple, TypeVar

from intervaltree import IntervalTree

from tdm.abstract.datamodel import AbstractSpan, AbstractTalismanSpan, AbstractTreeDocumentContent, NodeType
from .span import NullableSpan, TalismanSpan
from .tree import TreeNode

_TreeDocumentContent = TypeVar('_TreeDocumentContent', bound='TreeDocumentContent')


class TreeDocumentContent(TreeNode, AbstractTreeDocumentContent):
    __slots__ = (
        '_node_type', '_text', '_markup', '_original_text', '_hidden',
        '_text_references', '_concat_text_len', '_id2span', '_spans_tree'
    )

    DOCUMENT_SEPARATOR = '\n'  # Must be one whitespace character as methods implementations use this condition

    def __init__(self: _TreeDocumentContent, node_id: str, node_type: NodeType, text: str,
                 nodes: Optional[Iterable[_TreeDocumentContent]] = None,
                 markup: Optional[Dict[str, Any]] = None,
                 original_text: Optional[str] = None,
                 hidden: bool = False,
                 parent_func: Optional[Callable[[_TreeDocumentContent], _TreeDocumentContent]] = None,
                 text_translations: Optional[Dict[str, str]] = None):
        TreeNode.__init__(self, node_id, nodes, parent_func)

        self._node_type = node_type
        self._text = text
        self._text_translations = dict(text_translations) if text_translations is not None else {}
        self._markup = deepcopy(markup) if markup is not None else {}
        self._original_text = original_text
        self._hidden = hidden

        self._texts_references = [self._text]  # store references instead of concatted string to reduce memory usage
        self._id2span: Dict[str, NullableSpan] = {self._id: NullableSpan(0, len(self._text))}
        pointer = len(self._text) + len(TreeDocumentContent.DOCUMENT_SEPARATOR)
        for node in self._nodes:
            self._texts_references.extend(node._texts_references)
            self._id2span.update({node_id: span.shift(pointer) for node_id, span in node._id2span.items()})
            pointer += node._concat_text_len + len(TreeDocumentContent.DOCUMENT_SEPARATOR)
        self._concat_text_len = pointer - len(TreeDocumentContent.DOCUMENT_SEPARATOR)

        self._spans_tree = IntervalTree.from_tuples((s.start_idx, s.end_idx, n_id) for n_id, s in self._id2span.items() if s.length != 0)

    @property
    def type(self) -> NodeType:
        return self._node_type

    @property
    def text(self) -> str:
        return TreeDocumentContent.DOCUMENT_SEPARATOR.join(self._texts_references)

    def text_for(self, span: AbstractSpan) -> str:
        if not isinstance(span, AbstractTalismanSpan):
            span = self.talisman_span(span)
        if span.node_id == self._id:
            if span.end_idx > len(self._text):
                raise ValueError(f"Tree node {self._id} doesn't contain provided span: {span}")
            return self._text[span.start_idx:span.end_idx]
        if span.node_id not in self._id2node:
            raise ValueError(f"Tree node {self._id} doesn't contain provided span: {span}")
        return self._id2node[span.node_id].text_for(span)

    @property
    def node_text(self) -> str:
        return self._text

    def get_node_text_translation(self, lang: str) -> str:
        if lang not in self._text_translations:
            raise ValueError(f"Tree node {self._id} doesn't contain translation to {lang}")
        return self._text_translations[lang]

    @property
    def node_text_languages(self) -> Set[str]:
        return set(self._text_translations.keys())

    def with_node_translation(self: _TreeDocumentContent, lang: str, translation: str) -> _TreeDocumentContent:
        return self._with_changes(text_translations={**self._text_translations, lang: translation})

    def without_node_translation(self: _TreeDocumentContent, lang: str) -> _TreeDocumentContent:
        copied = dict(self._text_translations)
        if lang in copied:
            del copied[lang]
        return self._with_changes(text_translations=copied)

    @property
    def original_node_text(self) -> str:
        return self._original_text if self._original_text is not None else self._text

    def with_node_text(self: _TreeDocumentContent, text: str) -> _TreeDocumentContent:
        return self._with_changes(text=text)

    @property
    def markup(self) -> Dict[str, Any]:
        return deepcopy(self._markup)

    @property
    def is_hidden(self) -> bool:
        return self._hidden

    def hide(self: _TreeDocumentContent) -> _TreeDocumentContent:
        return self._with_changes(hidden=True, nodes=map(lambda node: node.hide(), self._nodes))

    def show(self: _TreeDocumentContent) -> _TreeDocumentContent:
        return self._with_changes(hidden=False, nodes=map(lambda node: node.show(), self._nodes))

    def talisman_spans(self, span: AbstractSpan) -> Tuple[AbstractTalismanSpan, ...]:
        if isinstance(span, AbstractTalismanSpan):
            if span.node_id == self._id:
                self._validate_talisman_span(span)
                return (span,)
            return self._id2node[span.node_id].talisman_spans(span)
        self._validate_span(span)
        overlapped_node_ids = self._spans_tree.overlap(span.start_idx, span.end_idx)
        result = []
        for _, _, node_id in overlapped_node_ids:
            node_span: NullableSpan = self._id2span[node_id]
            span_start = max(span.start_idx, node_span.start_idx) - node_span.start_idx
            span_end = min(span.end_idx, node_span.end_idx) - node_span.start_idx
            result.append(TalismanSpan(node_id, span_start, span_end))
        return tuple(result)

    def span(self, talisman_span: AbstractTalismanSpan) -> NullableSpan:
        self._validate_talisman_span(talisman_span)
        shift = self._id2span[talisman_span.node_id].start_idx
        return NullableSpan(talisman_span.start_idx + shift, talisman_span.end_idx + shift)

    def equal_structure(self: _TreeDocumentContent, other: _TreeDocumentContent) -> bool:
        if not isinstance(other, TreeDocumentContent):
            return False
        return self._node_type is other._node_type and self._text == other._text and super().equal_structure(other)

    def _validate_talisman_span(self, span: AbstractTalismanSpan) -> None:
        if span.node_id not in self._id2node:
            raise ValueError(f"provided span {span} is not contained in doc")
        if span.end_idx > len(self._id2node[span.node_id]._text):
            raise ValueError(f"Document {span.node_id} doesn't contain provided span: {span}")

    def _validate_span(self, span: AbstractSpan):
        if span.end_idx > self._concat_text_len:
            raise ValueError(f"Document {self._id} doesn't contain provided span: {span}")

    def _get_updated_constructor_params(self, **kwargs) -> dict:
        result = super()._get_updated_constructor_params(**kwargs)

        for param in ['node_type', 'original_text']:
            if param in kwargs:
                raise ValueError(f"Can't change {param} for node {self._id}")
        result['node_type'] = self._node_type

        if 'text' in kwargs:
            result['original_text'] = self._original_text if self._original_text is not None else self._text
            result['text'] = kwargs['text']
        else:
            result['original_text'] = self._original_text
            result['text'] = self._text

        result['text_translations'] = kwargs.get('text_translations', self._text_translations)

        result['markup'] = kwargs.get('markup', self._markup)
        result['hidden'] = kwargs.get('hidden', self._hidden)

        return result

    def __eq__(self, other):
        if not isinstance(other, TreeDocumentContent):
            return NotImplemented
        return TreeNode.__eq__(self, other) and self._node_type == other._node_type and self._text == other._text and \
            self._markup == other._markup and self._original_text == other._original_text and self._hidden == other._hidden



import uuid
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from enum import Enum
from functools import total_ordering
from typing import Any, Dict, Generic, Iterable, Optional, Tuple, TypeVar, Union

from tdm.abstract.datamodel.span import AbstractTalismanSpan


@total_ordering
class FactStatus(str, Enum):

    def __new__(cls, name: str, priority: int):
        obj = str.__new__(cls, name)
        obj._value_ = name
        obj.priority = priority
        return obj

    APPROVED = ("approved", 0)
    DECLINED = ("declined", 1)
    AUTO = ("auto", 2)
    HIDDEN = ("hidden", 3)
    NEW = ("new", 4)

    def __lt__(self, other: 'FactStatus'):
        if not isinstance(other, FactStatus):
            return NotImplemented
        return self.priority < other.priority


class FactType(str, Enum):
    PROPERTY = "property"
    RELATION = "relation"
    CONCEPT = "concept"
    VALUE = "value"


FactMetadata = Dict[str, Any]

_FactValue = TypeVar('_FactValue')
_AbstractFact = TypeVar('_AbstractFact', bound='AbstractFact')


class AbstractFact(Generic[_FactValue], metaclass=ABCMeta):
    __slots__ = ('_id', '_mention', '_fact_type', '_status', '_type_id', '_value', '_metadata')

    def __init__(self, id_: Optional[str], fact_type: FactType, status: FactStatus, type_id: str,
                 value: Optional[Union[_FactValue, Tuple[_FactValue, ...]]] = None,
                 mention: Iterable[AbstractTalismanSpan] = None, metadata: Optional[FactMetadata] = None):
        self._id = id_ or self.generate_id()
        self._mention = tuple(mention) if mention is not None else None
        self._fact_type = fact_type
        self._status = status
        self._type_id = type_id
        self._value = deepcopy(value)
        self._metadata = deepcopy(metadata)

    @property
    def id(self) -> str:
        return self._id

    @property
    def mention(self) -> Optional[Tuple[AbstractTalismanSpan, ...]]:
        return self._mention

    @property
    def fact_type(self) -> FactType:
        return self._fact_type

    @property
    def status(self) -> FactStatus:
        return self._status

    @property
    def type_id(self) -> str:
        return self._type_id

    @property
    def value(self) -> Union[_FactValue, Tuple[_FactValue, ...]]:
        return deepcopy(self._value)

    @property
    def has_value(self) -> bool:
        return bool(self._value)

    @property
    def metadata(self) -> FactMetadata:
        return deepcopy(self._metadata)

    @staticmethod
    def generate_id() -> str:
        return str(uuid.uuid4())

    def __eq__(self, other):
        if not isinstance(other, AbstractFact):
            return NotImplemented
        return self._id == other._id and self._mention == other._mention and self._fact_type == other._fact_type \
            and self._status == other._status and self._type_id == other._type_id and self._value == other._value

    def __hash__(self):
        return hash((self._id, self._mention, self._fact_type, self._status, self._type_id))

    @abstractmethod
    def with_changes(self: _AbstractFact, *, status: FactStatus = None, type_id: str = None,
                     value: Union[_FactValue, Tuple[_FactValue, ...]] = None,
                     mention: Tuple[AbstractTalismanSpan, ...] = None,
                     metadata: FactMetadata = None) -> _AbstractFact:
        pass

    def __repr__(self):
        return f"Fact({self._id}, {self._fact_type}, {self._status}, {self._type_id}, {self._value}, " \
               f"{self._metadata}, {self._mention})"


_SourceFactType = TypeVar('_SourceFactType', bound=AbstractFact)
_TargetFactType = TypeVar('_TargetFactType', bound=AbstractFact)


class AbstractLinkValue(Generic[_SourceFactType, _TargetFactType], metaclass=ABCMeta):
    def __init__(self, property_id: Optional[str], from_fact: _SourceFactType, to_fact: _TargetFactType):
        self.validate_slots(from_fact, to_fact)
        self._property_id = property_id
        self._from_fact = from_fact
        self._to_fact = to_fact

    @property
    def property_id(self) -> Optional[str]:
        return self._property_id

    @property
    def from_fact(self) -> _SourceFactType:
        return self._from_fact

    @property
    def to_fact(self) -> _TargetFactType:
        return self._to_fact

    def update_value(self, id2fact: Dict[str, AbstractFact]):
        return type(self)(self._property_id, id2fact[self._from_fact.id], id2fact[self._to_fact.id])

    @classmethod
    @abstractmethod
    def validate_slots(cls, source: _SourceFactType, target: _TargetFactType):
        pass

    def __eq__(self, other):
        if not isinstance(other, AbstractLinkValue):
            return NotImplemented
        return self._property_id == other._property_id and self._from_fact == other._from_fact and self._to_fact == self._to_fact

    def __hash__(self):
        return hash((self._property_id, self._from_fact, self._to_fact))

    def __repr__(self):
        return f"PropertyLinkValue({self._property_id}, {self.from_fact.id}, {self.to_fact.id})"
